/* ---------------------------------------------------------------- */
//
// AcroIO.m
// ADC
//
// Created by Christoph Birk on 2014-05-19
//
// 962-EN-4006:  6 channel analog input (voltage)
// 983-EN-4012: 12 channel digtal input/output
//
/* ---------------------------------------------------------------- */

#define DEBUG      0                   // debug level 

/* ---------------------------------------------------------------- */

#import  "AcroIO.h"

/* ---------------------------------------------------------------- */

@implementation AcroIO

/* ---------------------------------------------------------------- */
#pragma mark "Class"
/* ---------------------------------------------------------------- */

+ (NSString*)errorString:(int)err
{
  NSString *string=nil;
  
  switch (err) {
  case E_acro_error: string = @"Acromag: generic error"; break;
  case E_acro_ioerror: string = @"Acromag: incomplete response"; break;
  default:
    if (!string) string = [super errorString:err];
    break;
  }
  return string;
}

/* ---------------------------------------------------------------- */
#pragma mark "Overwrites"
/* ---------------------------------------------------------------- */

- (id)initWithHost:(const char*)hstr port:(int)pint online:(BOOL)on
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%s,%d,%d)\n",PREFUN,hstr,pint,on);
#endif

  self = [super initWithHost:hstr port:pint online:on];
  if (!self) return nil;

  terminator = '\0';                   // none
  
  return self;
}

/* ---------------------------------------------------------------- */

- (int)open
{
  int err;
#if (DEBUG > 2)
  fprintf(stderr,"%s:%p: count=%d\n",PREFUN,self,count);
#endif
  
  @synchronized(self) {
#if (DEBUG > 0)
    if (count == 0) fprintf(stderr,"%s(%p): %s\n",PREFUN,self,host);
#endif
    err = [super open];
  }
  return err;
}

/* ---------------------------------------------------------------- */

- (void)close
{
#if (DEBUG > 2)
  fprintf(stderr,"%s:%p: count=%d\n",PREFUN,self,count);
#endif

  @synchronized(self) {
    [super close];
#if (DEBUG > 0)
    if (count == 0) fprintf(stderr,"%s(%p): %s\n",PREFUN,self,host);
#endif
  }
}

/* ---------------------------------------------------------------- */
#pragma mark "Instance"
/* ---------------------------------------------------------------- */

- (int)send:(int)fc address:(int)addr value:(int)val response:(char*)res
{
  ssize_t n=0;
  char    message[32],buf[256];
#if (DEBUG > 2)
  fprintf(stderr,"%s(%d,%d,%d,%p)\n",PREFUN,fc,addr,val,res);
#endif

  int err = [self open]; if (err) return -err;

  if (![self mylock:lock_timeout]) {      // open/lock is more MT-efficient
    [self close];
    return -E_tcpip_lock;
  }
  
  sprintf(message,"%c%c%c%c%c%c%c%c%c%c%c%c",
          0x00,0x00,0x00,0x00,0x00,0x06,  // fixed length = 6 bytes
          0x00,                           // unit ID
          fc,                             // function code
          addr/0x100,addr%0x100,          // address
          val/0x100,val%0x100);           // data

  if (online) {
    assert(sock > 0);
    if (TCPIP_Send(sock,message,12)) n = -E_tcpip_send;
    if (!n) {                             // no error
      n = TCPIP_Recv(sock,buf,6,0);
      if (n == 6) {
        u_char bytes = (u_char)buf[5];    // bytes to come
        n = TCPIP_Recv(sock,buf,bytes,0);
        if (res && (n > 0)) memcpy(res,buf,n);
      } else {
        n = -E_tcpip_recv;
      }
    }
  } else {
    assert(sock == 0);
    [NSThread sleepForTimeInterval:0.020];
  }
  [lock unlock];
  [self close];

  return (int)n;                           // <0 is error code
}

/* --- */

- (int)send:(int)fc address:(int)addr value:(int)val
{
  return [self send:fc address:addr value:val response:NULL];
}

/* ---------------------------------------------------------------- */
#pragma mark "962-EN-4006"
/* ---------------------------------------------------------------- */

// Note: range: +-2.5 Volts (signal: 0.75..2.35 Volts)

- (int)getAnalog:(int)channel value:(int*)value
{
  char buf[256]="";                  // Clang wants an initialization
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d,%p)\n",PREFUN,channel,value);
#endif

  int err = [self open]; if (err) return err;
  
  int n = [self send:0x04 address:(0x0f+channel) value:0x0001 response:buf];
  if (online) {
    if      (n < 0) err = -n;
    else if (n < 5) err = E_acro_ioerror;
    else *value = 256 * (int)(u_char)buf[3] + (int)((u_char)buf[4]);
  }
  [self close];
  
  return err;
}

/* ---------------------------------------------------------------- */
#pragma mark "983-EN-4012"
/* ---------------------------------------------------------------- */

- (int)setCoil:(int)channel status:(int)status
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d,%d)\n",PREFUN,channel,status);
#endif

  int n = [self send:0x05 address:channel value:((status) ? 0xff00 : 0x0000)];
  if (online) {
    if (n < 0) return -n;
    if (n < 1) return E_acro_ioerror;
  }
  return 0;
}

/* ---------------------------------------------------------------- */

- (int)getCoil:(int)channel status:(int*)status
{
  char buf[256]="";                 // Clang wants an initialization
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d,%p)\n",PREFUN,channel,status);
#endif

  int n = [self send:0x02 address:channel value:0x0001 response:buf];
  if (online) {
    if (n < 0) return -n;
    if (n < 4) return E_acro_ioerror;
    else *status = (int)buf[3];
  }
  return 0;
}

/* ---------------------------------------------------------------- */

- (int)getInputs:(int*)value
{
  char res[256]="";
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,value);
#endif

  int n = [self send:0x02 address:0x0000 value:12 response:res];
  if (online) {
    if (n < 0) return -n;
    if (n < 5) return E_acro_ioerror;
    u_char byte1 = (u_char)res[3];  // res[3]: 0x01 .. 0x80 = inputs/coils 1..8
    u_char byte2 = (u_char)res[4];  // res[4]: 0x01 .. 0x08 = inputs/coils 9..12
    *value = (int)byte1 + 256*(int)byte2;
  }
  return 0;
}


/* ---------------------------------------------------------------- */

@end

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
