 /* ---------------------------------------------------------------- */
//
// ADC_Controller.m
// ADC
//
// Created by Christoph Birk on 2017-10-18
// Copyright (c) 2017 Christoph Birk. All rights reserved.
//
/* ---------------------------------------------------------------- */

#define DEBUG             1
#define UNIT_TEST         0      // 1:encoder conversion 2:test motion

/* ---------------------------------------------------------------- */

#import  "ADC_Controller.h"
#import  "AppDelegate.h"
#import  "Yutils.h"
#import  "PlotController.h"      // b0020
#import  "MagIO.h"
#import  "AcroIO.h"

#include "utils.h"

/* ---------------------------------------------------------------- */

enum lens_enum { LENS_A,LENS_B };
#define LENS_NAME(i)      (((i) == LENS_A) ? "A" : "B")

enum slide_enum { SLIDE_ADC,SLIDE_OTHER };
#define SLIDE_NAME(i)     (((i) == SLIDE_ADC) ? "ADC" : "Other")

#define DBE_AUTO_UPDATE   "dbe_auto_update"
#define DBE_ADC_STATE     "dbe_adc_state"
#define DBE_ENCODER_A     "dbe_lens_encoderA"
#define DBE_ENCODER_B     "dbe_lens_encoderB"
#define DBE_ENCODER(i)    (((i) == LENS_A) ? DBE_ENCODER_A : DBE_ENCODER_B)

#define DBE_OTHER_STATE   "dbe_other_state"

#define DBE_MIN_SIN(i)    (((i) == LENS_A) ? "dbe_min_sinA" : "dbe_min_sinB")
#define DBE_MAX_SIN(i)    (((i) == LENS_A) ? "dbe_max_sinA" : "dbe_max_sinB")
#define DBE_MIN_COS(i)    (((i) == LENS_A) ? "dbe_min_cosA" : "dbe_min_cosB")
#define DBE_MAX_COS(i)    (((i) == LENS_A) ? "dbe_max_cosA" : "dbe_max_cosB")

#define DBE_PHASE(i)      (((i) == LENS_A) ? "dbe_phaseA" : "dbe_phaseB")
#define DEF_PHASE(i)      (((i) == LENS_A) ? 45.0 : 90.0)  // +44.5,-178.0 b0054

enum adc_errors_enum {
  E_error=1,
  E_notInit,
  E_timeout,
  E_last
};

static const float pulseRange=0.7f,findRange=10.0f; 

/* ---------------------------------------------------------------- */

static float getAngle(int s,float v0s,float a0s,int c, float v0c,float a0c,float p)
{
  double sv = ((double)s - v0s)/a0s;
  double cv = ((double)c - v0c)/a0c;
  
  return modneg(p+RADS*atan2(sv,cv),360);  // norm={-180..+180}
}

/* ---------------------------------------------------------------- */

@implementation ADC_Controller
{
  AppDelegate *app;
  volatile BOOL   initialized,control_state,update_state;
  volatile BOOL   lensA_moving,lensB_moving;
  volatile BOOL   logging,calibrating;
  volatile BOOL   adc_moving,other_moving;
  volatile time_t next_update,interval_update;
  float           lensTargetA,lensTargetB;
  int             adcState,otherState;
#if (UNIT_TEST > 0)
  BOOL unit_stop;
#endif
  // Lens encoder calibration
  float  v0sinA,a0sinA,v0cosA,a0cosA,phaseA;
  float  v0sinB,a0sinB,v0cosB,a0cosB,phaseB;
  int    cal_min_sin,cal_max_sin,cal_min_cos,cal_max_cos;
  float  cal_phase;
  double calStartTime;
  PlotController *plotCircle,*plotSinCos,*plotAngle;
  // TCSIS
  MagIO *tcs;
  float elevation,rotator;
  // Acromag
  AcroIO *acroA,*acroD;
  float  lensEncoderA,lensEncoderB;
  int    adcLimitIn,adcLimitOut,otherLimitIn,otherLimitOut;
  NSLock *adcLock,*otherLock,*lensLockA,*lensLockB;
}

@synthesize control_state,update_state,adcState,but_run;

/* ---------------------------------------------------------------- */
#pragma mark "Class"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

+ (NSString*)errorString:(int)err
{
  NSString *string=nil;
  
  switch (err) {
  case E_error:   string = @"generic error"; break;
  case E_notInit: string = @"not initialized"; break;
  case E_timeout: string = @"motion timed out"; break;
  default:
    if (!string) string = [MagIO errorString:err];
    if (!string) string = [AcroIO errorString:err];
    break;
  }
  if (!string) string = [NSString stringWithFormat:@"unknown error (%d)",err];

  return string;
}

/* ---------------------------------------------------------------- */
#pragma mark "NSWindowController"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (id)initWithWindow:(NSWindow*)window
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p): %p\n",PREFUN,window,self);
#endif

  if (!(self = [super initWithWindow:window])) return nil;
  
  elevation = 90.0f;
  rotator   =  0.0f;
  
  update_state = YprefGetInt(DBE_AUTO_UPDATE,1,NO);
  
  adcLock   = [[NSLock alloc] init];
  lensLockA = [[NSLock alloc] init];
  lensLockB = [[NSLock alloc] init];
  otherLock = [[NSLock alloc] init];

  srand48(time(NULL));

  const PlotColor myColors[3] = {{0.65,0.0,0.0},{ 0.0,0.0,0.65},{0.0,0.65,0.0}};
  plotSinCos = [[PlotController alloc] initWithWindowNib:@"Plot"
                                        context:0 title:"Calibration/SinCos"
                                        unit:"" N:2 colors:myColors];

  plotCircle = [[PlotController alloc] initWithWindowNib:@"Plot"
                                       context:0 title:"Calibration/Circle"
                                       unit:"" N:1 colors:myColors];

  plotAngle = [[PlotController alloc] initWithWindowNib:@"Plot"
                                       context:0 title:"Calibration/Angle"
                                       unit:"dg" N:1 colors:&myColors[2]];

  return self;
}

/* ---------------------------------------------------------------- */

- (void)awakeFromNib
{
#if (DEBUG > 0)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif

#if 0 // NSWindowController has an "empty" implementation
  if ([[super class] instancesRespondToSelector:@selector(awakeFromNib)]) {
    fprintf(stderr,"%s:%p: super-class=%p\n",PREFUN,self,[super class]);
    [super awakeFromNib];
  }
#endif

  assert([NSThread isMainThread]);
  app = (AppDelegate*)[NSApp delegate]; // b10052

  [chk_update setState:((update_state) ? NSControlStateValueOn : NSControlStateValueOff)];

  assert([NSThread isMainThread]);     // timer on main thread
  [NSTimer scheduledTimerWithTimeInterval:1.0 target:self
           selector:@selector(fireTimer:) userInfo:nil repeats:YES];

#if (UNIT_TEST == 1)
  [self reloadPreferences];
#if 1
  for (float x=0; x<360; x+=30) {
    lensEncoderA = x;
    [self updateLens:LENS_A encoder:YES];
    printf("x=%f, lensEncoderA=%f\n",x,lensEncoderA);
    assert(fabs(modneg(x-lensEncoderA,360)) < 0.01);
  }
#endif
#if 1
  for (float x=0; x<360; x+=30) {
    lensEncoderB = x;
    [self updateLens:LENS_B encoder:YES];
    printf("x=%f, lensEncoderB=%f\n",x,lensEncoderB);
    assert(fabs(modneg(x-lensEncoderB,360)) < 0.01);
  }
#endif
#endif // UNIT TEST 1

#if (UNIT_TEST == 2)
  dispatch_async(GLOBAL_QUEUE,^(void) { float a=0,b=0;
    double a1=0,a2=0,an=0,b1=0,b2=0,bn=0; int fa=0,fb=0;
    while (!initialized) { Ysleep(350); }
    printf("UNIT_TEST=%d\n",UNIT_TEST);
    unit_stop = FALSE;
    double testTime = 300+walltime();
    do {
      if (!lensA_moving) {
        if (a > 0) {
          double d = fabs(a-lensEncoderA); a1 += d; a2 += d*d; an += 1.0;
          double ave = a1/an;
          double sig = sqrt(a2/an-ave*ave);
          printf("A: %.2f (%.1f,%.1f,%.0f)\n",d,ave,sig,an);
          if (d > 1.5*pulseRange) { char buf[128];
            fa++; sprintf(buf,"A: %d/%.0f fails",fa,an);
            [main_logger append:buf]; fprintf(stderr,"%s\n",buf);
          }
        }
        Ysleep(2000); a = 360.0*drand48();
        dispatch_sync(MAIN_QUEUE,^(void) { char buf[128];
          sprintf(buf,"%.1f",a); YeditSet(edit_encA,buf);
          [self action_lens:edit_encA];
        });
      }
      if (!lensB_moving) {
        if (b > 0) {
          double d = fabs(b-lensEncoderB); b1 += d; b2 += d*d; bn += 1.0;
          double ave = b1/bn;
          double sig = sqrt(b2/bn-ave*ave);
          printf("B: %.2f (%.1f,%.1f,%.0f)\n",d,ave,sig,bn);
          if (d > 1.5*pulseRange) { char buf[128];
            fb++; sprintf(buf,"B: %d/%.0f fails",fb,bn);
            [main_logger append:buf]; fprintf(stderr,"%s\n",buf);
          }
        }
        Ysleep(2000); b = 360.0*drand48();
        dispatch_sync(MAIN_QUEUE,^(void) { char buf[128];
          sprintf(buf,"%.1f",b); YeditSet(edit_encB,buf);
          [self action_lens:edit_encB];
        });
      }
      Ysleep(350);
      if (unit_stop) break;
    } while (walltime() < testTime);
    printf("UNIT_TEST=%d done\n",UNIT_TEST);
  });
#endif // UNIT TEST 2
}

/* ---------------------------------------------------------------- */

- (void)windowWillClose:(NSNotification*)note 
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p): object=%p\n",PREFUN,note,note.object);
#endif

  if (calibrating) [self action_calibrate:nil];   // stop calibration
  
  if (note.object == [app panel_calibrate]) return;
  
  YprefPutInt(DBE_AUTO_UPDATE,update_state,NO);
  
  YprefPutInt(DBE_ADC_STATE,adcState,use_shared); // store last known
  YprefPutInt(DBE_OTHER_STATE,otherState,use_shared);
  YprefPutDouble(DBE_ENCODER_A,lensEncoderA,use_shared);
  YprefPutDouble(DBE_ENCODER_B,lensEncoderB,use_shared);
}

/* ---------------------------------------------------------------- */
#pragma mark "Instance"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (void)reloadCalibration:(int)lens
{
  char buf[128]="";
#if (DEBUG > 0)
  fprintf(stderr,"%s(%s)\n",PREFUN,LENS_NAME(lens));
#endif
  assert(!calibrating);

  if (lens == LENS_A) { // updated default values b10054
    cal_min_sin = YprefGetInt(DBE_MIN_SIN(LENS_A),40790,use_shared); // 40721
    cal_max_sin = YprefGetInt(DBE_MAX_SIN(LENS_A),58030,use_shared); // 57986
    v0sinA = (cal_max_sin+cal_min_sin)/2.0;
    a0sinA = (cal_max_sin-cal_min_sin)/2.0;
    cal_min_cos = YprefGetInt(DBE_MIN_COS(LENS_A),40797,use_shared); // 40784
    cal_max_cos = YprefGetInt(DBE_MAX_COS(LENS_A),57941,use_shared); // 57956
    v0cosA = (cal_max_cos+cal_min_cos)/2.0;
    a0cosA = (cal_max_cos-cal_min_cos)/2.0;
    phaseA = YprefGetDouble(DBE_PHASE(LENS_A),DEF_PHASE(LENS_A),use_shared);
    sprintf(buf,"calibPref(A): vSin=%7.1f aSin=%7.1f vCos=%7.1f aCos=%7.1f phase=%+.1f",
            v0sinA,a0sinA,v0cosA,a0cosA,phaseA);
  } else
  if (lens == LENS_B) {
    cal_min_sin = YprefGetInt(DBE_MIN_SIN(LENS_B),40982,use_shared); // 40860
    cal_max_sin = YprefGetInt(DBE_MAX_SIN(LENS_B),57796,use_shared); // 57798
    v0sinB = (cal_max_sin+cal_min_sin)/2.0;
    a0sinB = (cal_max_sin-cal_min_sin)/2.0;
    cal_min_cos = YprefGetInt(DBE_MIN_COS(LENS_B),40851,use_shared); // 40787
    cal_max_cos = YprefGetInt(DBE_MAX_COS(LENS_B),57659,use_shared); // 57675
    v0cosB = (cal_max_cos+cal_min_cos)/2.0;
    a0cosB = (cal_max_cos-cal_min_cos)/2.0;
    phaseB = YprefGetDouble(DBE_PHASE(LENS_B),DEF_PHASE(LENS_B),use_shared);
    sprintf(buf,"calibPref(B): vSin=%7.1f aSin=%7.1f vCos=%7.1f aCos=%7.1f phase=%+.1f",
            v0sinB,a0sinB,v0cosB,a0cosB,phaseB);
  } else {
    assert(0);
  }
  [main_logger append:buf];
  
  dispatch_async(MAIN_QUEUE,^(void) {  // b0029
    if (lens == [pop_lens indexOfSelectedItem]) { char buf[128];
      cal_phase = YprefGetDouble(DBE_PHASE(lens),DEF_PHASE(lens),use_shared);
      sprintf(buf,"%.1f",cal_phase);
      YeditSet(edit_phase,buf);
    }
  });
}

/* --- */

- (void)reloadPreferences
{
  char buf[128];
#if (DEBUG > 0)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif
  
  [self reloadCalibration:LENS_A];
  [self reloadCalibration:LENS_B];

  interval_update = YprefGetInt(DBE_INTERVAL,DEF_INTERVAL,use_shared);
  sprintf(buf,"interval= %ld , mode= %d",interval_update,[app prism_mode]);
  [main_logger append:buf];            // b0043
  
  if (update_state) {
    @synchronized(self) { next_update = 0; }
  } else {
    [self updateLensTarget:control_state];
  }
}

/* ---------------------------------------------------------------- */

- (void)fireTimer:(NSTimer*)theTimer
{
#if (DEBUG > 2)
  fprintf(stderr,"%s(%p): init=%d, inter=%ld, auto=%d\n",PREFUN,theTimer,
          initialized,interval_update,update_state);
#endif

  if (!initialized) return;
  
  if (calibrating) {                     // cleanup plots
    if (walltime()-calStartTime > 120) {
      if (drand48() < 0.35) [plotCircle removeData:5];
      if (drand48() < 0.35) [plotSinCos removeData:5];
      if (drand48() < 0.35) [plotAngle  removeData:5];
    }
    return;
  }

  time_t ut = time(NULL);
  if (ut >= next_update) {
    if (update_state) {                  // Auto ?
      [but_update setTitle:@"Update"];
      if (lensA_moving || lensB_moving) { char buf[128];
        sprintf(buf,"%s: delay lens correction (%d,%d)",PREFUN,
          lensA_moving,lensB_moving);    // b0050
        [main_logger append:buf];
      } else {
        dispatch_async(GLOBAL_QUEUE,^(void) {
          [self update:control_state];
        }); 
        @synchronized(self) { next_update = ut + interval_update; }
      }
    }
  } else {                               // display countdown b0050
    if (update_state) {
      NSString *s = [NSString stringWithFormat:@"%ld",next_update-ut];
      [but_update setTitle:s];
    }
  }
}

/* ---------------------------------------------------------------- */

- (void)showError:(NSString*)text error:(int)err
{
  NSString *string=nil;
  
  string = [NSString stringWithFormat:@"%@, err=%d\r%@",text,
             err,[[self class] errorString:err]];
  [main_logger message:[string UTF8String] level:2 time:60 file:YES];
}

/* ---------------------------------------------------------------- */
#pragma mark "Actions"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (IBAction)action_telescope:(id)sender
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif
  assert([NSThread isMainThread]);

  if (sender == chk_update) {          // automatic update
    update_state = ([sender state] == NSControlStateValueOn) ? YES : NO;
    [but_update setTitle:@"Update"];
    [edit_elevation setEditable:!update_state];
    [edit_elevation setSelectable:!update_state];
    [edit_rotator setEditable:!update_state];
    [edit_rotator setSelectable:!update_state];
    [app updateRunMode]; // b0037
    @synchronized(self) { next_update = 0; }  // ASAP
  } else
  if (sender == but_update) {          // manual update
    [but_update setTitle:@"Update"];
    dispatch_async(GLOBAL_QUEUE,^(void) {
      int err = [self openAcros];      // b0038
      if (!err) {
        if (!err) err = [self updateLens:LENS_A encoder:YES];
        if (!err) err = [self updateLens:LENS_B encoder:YES];
        if (!err) err = [self updateSlides:YES];
        if (!err) [self update:control_state];
        Ysleep(500);
        [self closeAcros];
      }
      if (err) [self showError:@"update failed" error:err];
    });
#ifndef NDEBUG
  } else
  if (sender == edit_elevation) {
    elevation = [sender floatValue];
    [self updateLensTarget:control_state];
  } else
  if (sender == edit_rotator) {
    rotator = [sender floatValue];
    [self updateLensTarget:control_state];
#endif
  } else {
   assert(0);
  }
}

/* ---------------------------------------------------------------- */

- (IBAction)action_slide:(id)sender
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif
  assert([NSThread isMainThread]);

  int position = (int)[sender indexOfSelectedItem];
  
  if (sender == popup_adc) {
    dispatch_async(GLOBAL_QUEUE,^(void) {
      int err = [self moveSlide:SLIDE_ADC to:position];
      if (err) [self showError:@"moving ADC failed" error:err];
      [app updateRunMode]; // b0037
    });
  } else
  if (sender == popup_other) {
    dispatch_async(GLOBAL_QUEUE,^(void) {
      int err = [self moveSlide:SLIDE_OTHER to:position];
      if (err) [self showError:@"moving Other failed" error:err];
    });
  } else {
    assert(0);
  }
}

/* ---------------------------------------------------------------- */

- (IBAction)action_lens:(id)sender
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif
  assert([NSThread isMainThread]);
  // edit angle edit not disabled -- assert(!calibrating);
  if (calibrating) return;

  float pos = (float)modneg([sender floatValue],360); // norm={-180..+180} bugfix b10055
  dispatch_async(GLOBAL_QUEUE,^(void) { int err;
    err = [self moveLens:((sender == edit_encA) ? LENS_A : LENS_B) to:pos];
    if (err) [self showError:@"moving lens failed" error:err];
  });
}

/* ---------------------------------------------------------------- */

- (IBAction)action_control:(id)sender
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif
  assert([NSThread isMainThread]);

  if (sender == popup_control) {
    control_state = ([sender indexOfSelectedItem] == 0) ? NO : YES;
  } else {
    control_state = [sender intValue];
    [popup_control selectItemAtIndex:control_state];
  }
  [app updateRunMode]; // b0037

  [self enableGUI:YES];
  if (control_state) [self moveLensToTarget]; // b0018
}

/* ---------------------------------------------------------------- */

- (IBAction)action_calibrate:(id)sender
{
  char buf[128];
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif
  assert([NSThread isMainThread]);
  
  int lens = (int)[pop_lens indexOfSelectedItem];

  if (sender == pop_lens) {            // b0029
    cal_phase = YprefGetDouble(DBE_PHASE(lens),DEF_PHASE(lens),use_shared);
    sprintf(buf,"%.1f",cal_phase);
    YeditSet(edit_phase,buf);
    return;
  }

  if (sender == edit_phase) {
    cal_phase = [edit_phase floatValue];
#ifndef SIM_ONLY
    YprefPutDouble(DBE_PHASE(lens),cal_phase,use_shared);
#endif
    if (!calibrating) {
      [self reloadCalibration:lens];
      [self updateLens:lens encoder:YES];
      if (lens==LENS_A) sprintf(buf,"+%.1f",lensEncoderA);
      else              sprintf(buf,"+%.1f",lensEncoderB);
      YeditSet(edit_angle,buf);        // b0036
    }
    return;
  }
  
  if (sender == but_plus) {
    int err = [self moveLens:lens dir:(+1)];
    if (!err) {
      [sender setEnabled:NO];
      [but_minus setEnabled:YES];
      if (!calibrating && !logging) {
        [self performSelectorInBackground:@selector(run_log:) withObject:@(lens)];
      }
    } else {
      [self showError:@"moving lens failed" error:err];
    }
    return;
  }
  if (sender == but_minus) {
    int err = [self moveLens:lens dir:(-1)];
    if (!err) {
      if (!calibrating && !logging) {
        [self performSelectorInBackground:@selector(run_log:) withObject:@(lens)];
      }
      [sender setEnabled:NO];
      [but_plus setEnabled:YES];
    } else {
      [self showError:@"moving lens failed" error:err];
    }
    return;
  }
  if (sender == but_stop) {
    int err = [self moveLens:lens dir:0];
    if (err) [self showError:@"stopping lens failed" error:err];
    @synchronized(self) { logging = FALSE; }
    [but_plus setEnabled:YES];
    [but_minus setEnabled:YES];
#if (UNIT_TEST == 2)
    unit_stop = TRUE;
#endif
    return;
  }
  
  // maybe 'nil' -- assert(sender == but_calibrate);
  if (calibrating) {                   // -> stop
    calibrating = FALSE;
    return;
  }
  
  calibrating = TRUE;                  // -> start
  if (control_state) [self action_control:@(0)];  // turn-off control
  [pop_lens setEnabled:NO];
  [but_calibrate setTitle:@"Stop Calibration"];
  cal_phase = [edit_phase floatValue];
  [self performSelectorInBackground:@selector(run_calibration:) withObject:@(lens)];
}

/* ---------------------------------------------------------------- */

- (IBAction)action_run:(id)sender      // b0040
{
  assert([NSThread isMainThread]);
  
  int adc = ([[sender title] compare:@"Run"] == NSOrderedSame) ? 1 : 0;
  if (adcState != adc) {               // insert ADC
    [popup_adc selectItemAtIndex:adc];
    [self action_slide:popup_adc];
  }
  BOOL run = ([[sender title] compare:@"Run"] == NSOrderedSame) ? YES : NO;
  if (update_state != run) {           // auto-update telescope
    [chk_update setState:(run ? NSControlStateValueOn : NSControlStateValueOff)];
    [self action_telescope:chk_update];
  }
  if (control_state != run) {          // auto control
    if (run)  [self action_control:@(1)];
    else      [self action_control:@(0)];
  }
  if (run) {
    @synchronized(self) { next_update = 0; }
  }
}

/* ---------------------------------------------------------------- */
#pragma mark "Update/Enable"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (void)enableGUI:(BOOL)enable
{
  dispatch_async(MAIN_QUEUE,^(void) {
    if (adc_moving) {
      [prog_adc startAnimation:self];
      [popup_adc setEnabled:NO];
    } else {
      [prog_adc stopAnimation:self];
      [popup_adc setEnabled:enable];
    }
    if (other_moving) {
      [prog_other startAnimation:self];
      [popup_other setEnabled:NO];
    } else {
      [prog_other stopAnimation:self];
      [popup_other setEnabled:enable];
    }
    [popup_control setEnabled:enable];
    [edit_encA setEnabled:(!lensA_moving && enable && !control_state)];
    [edit_encB setEnabled:(!lensB_moving && enable && !control_state)];
    if (lensA_moving) [prog_lensA startAnimation:self];
    else              [prog_lensA stopAnimation:self];
    if (lensB_moving) [prog_lensB startAnimation:self];
    else              [prog_lensB stopAnimation:self];
    int any = adc_moving + other_moving + lensA_moving + lensB_moving;
    [but_run setEnabled:(enable && !any)]; 
  });
}

/* ---------------------------------------------------------------- */

- (int)updateTCS:(BOOL)fromHardware
{
  int err=0;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d)\n",PREFUN,fromHardware);
#endif
  
  if (fromHardware) { // TODO: keep open if interval too short
    if (tcs.online) { float r,e;
      err = [tcs getAirmass:NULL rotangle:NULL rotatore:&r
                 azimuth:NULL elevation:&e para:NULL];
      if (!err) { rotator = r; elevation = e; }
    } else {
#ifdef SIM_ONLY
      static double sign=0.0; // TESTING
      if (elevation > 89.0) sign = -1.0;
      if (elevation < 30.0) sign =  1.0;
      elevation += sign*drand48();
      rotator = modulo(rotator+2.0*drand48(),0,360.0);
#endif
      Ysleep(250);
    }
  }
  
  dispatch_async(MAIN_QUEUE,^(void) { char buf[128];
    sprintf(buf,"%.2f",elevation); YeditSet(edit_elevation,buf);
    sprintf(buf,"%.2f",rotator); YeditSet(edit_rotator,buf);
  });
  
  return err;
}

/* ---------------------------------------------------------------- */

- (int)updateLimits:(BOOL)fromHardware
{
  int err=0;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%d)\n",PREFUN,fromHardware);
#endif

  if (fromHardware) { int v=0;
    err = [acroD getInputs:&v];
    if (!err && acroD.online) {
      otherLimitOut = (v & 0x10) ? 1 : 0; // bit-4
      otherLimitIn  = (v & 0x20) ? 1 : 0; // bit-5
      adcLimitOut   = (v & 0x40) ? 1 : 0; // bit-6
      adcLimitIn    = (v & 0x80) ? 1 : 0; // bit-7
    }
  }
  if      (!adcLimitIn &&  adcLimitOut) adcState =  0;
  else if ( adcLimitIn && !adcLimitOut) adcState =  1;
  else                                  adcState = -1;
  if      (!otherLimitIn &&  otherLimitOut) otherState =  0;
  else if ( otherLimitIn && !otherLimitOut) otherState =  1;
  else                                      otherState = -1;

  dispatch_async(MAIN_QUEUE,^(void) { // 'IB' dependency
    NSColor *off = [NSColor controlColor];  // b0040
    NSColor *on  = [NSColor systemGreenColor];
    col_adcIn.color    = (adcLimitIn)    ? on : off;
    col_adcOut.color   = (adcLimitOut)   ? on : off;
    col_otherIn.color  = (otherLimitIn)  ? on : off;
    col_otherOut.color = (otherLimitOut) ? on : off;
  });
  
  return err;
}

/* ---------------------------------------------------------------- */

- (int)updateSlides:(BOOL)fromHardware
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%d)\n",PREFUN,fromHardware);
#endif

  int err = [self updateLimits:fromHardware];
  if (!err) { char buf[128];           // log state/limits b0051
    sprintf(buf,"adcState= %d (%d,%d), otherState= %d (%d,%d)",
            adcState,adcLimitOut,adcLimitIn,otherState,otherLimitOut,otherLimitIn);
    [main_logger append:buf];
  }

  dispatch_async(MAIN_QUEUE,^(void) {
    [popup_adc selectItemAtIndex:adcState];
    [popup_other selectItemAtIndex:otherState];
  });
  
  return err;
}

/* ---------------------------------------------------------------- */

- (void)updateLensTarget:(BOOL)moveIt
{
  char buf[128];
#if (DEBUG > 0)
  fprintf(stderr,"%s(%d)\n",PREFUN,moveIt);
#endif
  
  double z = 90.0-elevation;
  double y = fmin(1.0,tan(z/RADS) / 1.665); // 59.01 deg
  double a = acos(y)*RADS;
  double r = 90.0-rotator;
  switch ([app prism_mode]) {
  case 0: default:                    // 'IB' dependency
    lensTargetA = modneg(r+180-a,360); //b0039
    lensTargetB = modneg(r-180+a,360);
    break;
  case 1:
    lensTargetA = modneg(r+90,360);
    lensTargetB = modneg(r-90,360);
    break;
  case 2:
    lensTargetA = modneg(r+a,360);    // b0039
    lensTargetB = modneg(r-a,360);
    break;
  }
  
  dispatch_async(MAIN_QUEUE,^(void) {
    [edit_targetA setStringValue:[NSString stringWithFormat:@"%+.1f",lensTargetA]];
    [edit_targetB setStringValue:[NSString stringWithFormat:@"%+.1f",lensTargetB]];
  });
  sprintf(buf,"%s(%d): ele= %6.3f rot= %8.3f tgA= %7.2f  tgB= %7.2f",PREFUN,moveIt,
          elevation,rotator,lensTargetA,lensTargetB);
  [main_logger append:buf];

  if (!calibrating && moveIt) [self moveLensToTarget]; // b0018
}

/* ---------------------------------------------------------------- */

- (int)getEncoder:(int)lens sin:(int*)s cos:(int*)c
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%s,%p,%p)\n",PREFUN,LENS_NAME(lens),s,c);
#endif

  int err = [acroA open];
  if (!err) {
    if (lens == LENS_A) {
      if (!err) err = [acroA getAnalog:0 value:s];
      if (!err) err = [acroA getAnalog:1 value:c];
    } else
    if (lens == LENS_B) {
      if (!err) err = [acroA getAnalog:2 value:s];
      if (!err) err = [acroA getAnalog:3 value:c];
    } else {
      assert(0);
    }
    [acroA close];
  }
  return err;
}

/* --- */

- (int)updateLens:(int)lens encoder:(BOOL)fromHardware
{
  int  err=0;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d)\n",PREFUN,fromHardware);
#endif

  if (fromHardware) { double sv,cv;
    assert(!calibrating);
    if (lens == LENS_A) {            // Simulator preset
      sv = v0sinA + a0sinA*sin((lensEncoderA-phaseA)/RADS);
      cv = v0cosA + a0cosA*cos((lensEncoderA-phaseA)/RADS);
    } else {
      assert(lens == LENS_B);
      sv = v0sinB + a0sinB*sin((lensEncoderB-phaseB)/RADS);
      cv = v0cosB + a0cosB*cos((lensEncoderB-phaseB)/RADS);
    }
    int s = (int)my_round(sv,0);
    int c = (int)my_round(cv,0);
    err = [self getEncoder:lens sin:&s cos:&c];
    if (!err) {
      if (lens == LENS_A) {
        lensEncoderA = getAngle(s,v0sinA,a0sinA,c,v0cosA,a0cosA,phaseA);
      } else {
        lensEncoderB = getAngle(s,v0sinB,a0sinB,c,v0cosB,a0cosB,phaseB);
      }
    }
  }

  dispatch_async(MAIN_QUEUE,^(void) { char buf[128];
    if (lens == LENS_A) {
      sprintf(buf,"%+.1f",lensEncoderA); YeditSet(edit_encA,buf);
    } else {
      assert(lens == LENS_B);
      sprintf(buf,"%+.1f",lensEncoderB); YeditSet(edit_encB,buf);
    }
  });
  
  return err;
}

/* ---------------------------------------------------------------- */

- (void)update:(BOOL)moveLenses
{
#if (DEBUG > 0)
  fprintf(stderr,"%s:(%d)\n",PREFUN,moveLenses);
#endif
  
  int err = [self updateTCS:YES];
  if (err) {
    [self showError:@"connecting to TCS failed" error:err];
  } else {
    [self updateLensTarget:moveLenses];
  }
}

/* ---------------------------------------------------------------- */

- (void)updatePlot:(int)lens sin:(int)s cos:(int)c
{
  char buf[128];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d,%d,%d)\n",PREFUN,lens,s,c);
#endif

  sprintf(buf,"lens(%s): sin= %5d cos= %5d",LENS_NAME(lens),s,c);
  [main_logger message:buf file:YES];

  double t = walltime() - calStartTime;
  [plotSinCos addData:0 x:t y:s redraw:NO];
  [plotSinCos addData:1 x:t y:c redraw:YES];
  [plotCircle addData:0 x:c y:s redraw:YES];
}

/* --- */

- (void)updatePlot:(int)lens sin:(int)s cos:(int)c angle:(float)a
{
  if ((a >= -180) && (a < 360)) {
   double t = walltime() - calStartTime;
   [plotAngle addData:0 x:t y:modulo(a,0,360) redraw:YES]; // norm={0..360}
  }
  [self updatePlot:lens sin:s cos:c];
}

/* ---------------------------------------------------------------- */
#pragma mark "Move"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (int)openAcros                       // b0038
{
  int err = [acroD open];
  if (!err) {
    err = [acroA open];
    if (err) [acroD close];
  }
  return err;                          // both open or both closed
}

/* --- */

- (void)closeAcros
{
  [acroD close];
  [acroA close];
}

/* ---------------------------------------------------------------- */

- (void)preset:(int)slide ilim:(int)i olim:(int)o
{
  if (slide == SLIDE_ADC) {
    adcLimitIn = i; adcLimitOut = o;
  } else {
    otherLimitIn = i; otherLimitOut = o;
  }
}

/* --- */

- (int)state:(int)slide
{
  return (slide == SLIDE_ADC) ? adcState : otherState;
}

/* --- */

- (int)moveSlide:(int)slide to:(int)pos   // b0029
{
  char buf[128];
#if (DEBUG > 0)
  fprintf(stderr,"%s(%d,%d)\n",PREFUN,slide,pos);
#endif

  if (slide == SLIDE_ADC) {
    [adcLock lock]; adc_moving = TRUE;
  } else {
    [otherLock lock]; other_moving = TRUE;
  }
  [self enableGUI:YES];
  sprintf(buf,"%s(%s,%d)",PREFUN,SLIDE_NAME(slide),pos);
  [main_logger append:buf];
  
  int err = [acroD open];
  if (!err) { double timeout = 10+walltime();
    if (!acroD.online) [self preset:slide ilim:0 olim:0]; // Simulator preset
    int coil = (slide == SLIDE_ADC) ? 0 : 1;
    while (walltime() < timeout) {
      err = [self updateLimits:YES]; if (err) break;
      if ([self state:slide] == pos) break;
      err = [acroD setCoil:coil status:(pos==1) ? 1 : 0]; if (err) break;
      Ysleep(1000);
      if (!acroD.online) {         // Simulator preset
        [self preset:slide ilim:((pos) ? 1 : 0) olim:((pos) ? 0 : 1)];
      }
    } // endwhile(timeout)
    if (!err) [self updateLimits:YES];
    [acroD close];
    if (!err) {
      if ([self state:slide] != pos) err = E_timeout;
    }
  }
  [self updateSlides:NO];

  if (slide == SLIDE_ADC) {
    adc_moving = FALSE;  [adcLock unlock];
  } else {
    other_moving = FALSE;  [otherLock unlock];
  }
  [self enableGUI:YES];
  
  return err;
}

/* ---------------------------------------------------------------- */

- (int)moveLens:(int)lens dir:(float)dir
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d,%.2f)\n",PREFUN,lens,dir);
#endif

  int err = [acroD open];
  if (!err) {
    int b1 = (lens == LENS_A) ? 10 : 8;
    int b2 = (lens == LENS_A) ? 11 : 9;
    if (dir > 0) {
      if (!err) err = [acroD setCoil:b1 status:0];
      if (!err) err = [acroD setCoil:b2 status:1];
    } else
    if (dir < 0) {
      if (!err) err = [acroD setCoil:b1 status:1];
      if (!err) err = [acroD setCoil:b2 status:0];
    } else {
      if (!err) err = [acroD setCoil:b1 status:0];
      if (!err) err = [acroD setCoil:b2 status:0];
    }
    [acroD close];
  }
  
  return err;
}

/* ---------------------------------------------------------------- */

- (float)encoder:(int)lens
{
  return (lens == LENS_A) ? lensEncoderA : lensEncoderB;
}

/* --- */

- (int)find:(float)pos lens:(int)lens
{
  int err=0;
  
  double timeout = 30+walltime();
  while (walltime() < timeout) {
    err = [self updateLens:lens encoder:YES]; if (err) break;
    double dist = modneg(pos-[self encoder:lens],360);
    if (fabs(dist) <= findRange) break;
    err = [self moveLens:lens dir:dist]; if (err) break;
    Ysleep(350);                // less than 6*80=480
    if (!acroA.online) {        // Simulator
      if (lens == LENS_A) lensEncoderA += dist/2.0f;
      else                lensEncoderB += dist/2.0f;
    }
  } // endwhile(!timeout)
  [self moveLens:lens dir:0];   // stop motor

  return err;
}

/* --- */

- (int)pulse:(float)pos lens:(int)lens
{
  int    err=0;
  double dist=0,slpt;
  char   buf[128];
  
  NSProcessInfo *info = NSProcessInfo.processInfo; // b0044
  double timeout = 10+walltime();      // was 15 b0046
  while (walltime() < timeout) {
    err = [self updateLens:lens encoder:YES]; if (err) break;
    dist = modneg(pos-[self encoder:lens],360);
    if (fabs(dist) <= pulseRange) break;
    id token = [info beginActivityWithOptions:NSActivityLatencyCritical reason:@(PREFUN)];
    err = [self moveLens:lens dir:dist];
    if (!err) {
      double t1 = walltime();
      slpt = fabs(dist)/20.0;          // bug-fix b0043
      [NSThread sleepForTimeInterval:slpt];
      [self moveLens:lens dir:0];      // stop motor
      double dt = walltime()-t1;
      sprintf(buf,"dist%s= %+.1f slpt= %.4f dt= %.4f",LENS_NAME(lens),dist,slpt,dt);
      [main_logger append:buf];        // b0045
      // NOTE: checked for oscillations 2019-07-15
      //       checked for "err=3" 2019-07-15 ... none
      //       after PFS run 2019-08-27 ... 1 per week
      //  moveLens:to:](X,
      //  distX=
      //  moveX=
      // found 10% "overshoots", but <1% oscillations
      // A: 1479 159 7
      // B: 1448 120 1
    }
    [info endActivity:token];
    if (err) break;
    Ysleep(650);  // wait for encoder sampling (was 500,600) b0046,b0047
    if (!acroA.online) {
      if (lens == LENS_A) lensEncoderA += drand48()*dist;
      else                lensEncoderB += drand48()*dist;
    }
  } // endwhile (time out)

  return err;
}

/* ---------------------------------------------------------------- */

- (int)moveLens:(int)lens to:(float)pos
{
  char buf[128];
#if (DEBUG > 0)
  fprintf(stderr,"%s(%d,%.3f)\n",PREFUN,lens,pos);
#endif
  assert(!calibrating);

  if (lens == LENS_A) {                  // semaphore
    [lensLockA lock]; lensA_moving = TRUE;
  } else {
    [lensLockB lock]; lensB_moving = TRUE;
  }
  sprintf(buf,"%s(%s,%.3f)",PREFUN,LENS_NAME(lens),pos);
  [main_logger append:buf];
  [self enableGUI:YES];

  int err = [self openAcros];            // b0038
  if (!err) {
    float move = modneg(pos-[self encoder:lens],360); // modneg() b0044
    if (!err) err = [self find:pos lens:lens];
    if (!err) Ysleep(960);               // more than 6*80=480 b0047
    if (!err) err = [self pulse:pos lens:lens];
    if (!err) Ysleep(500);
    if (!err) err = [self updateLens:lens encoder:YES];
    [self closeAcros];
    if (!err) {
      double dist = modneg(pos-[self encoder:lens],360); // {-180,+180}
      sprintf(buf,"move%s= %+.1f  dist= %+.1f",LENS_NAME(lens),move,dist); // sign b0051
      [main_logger append:buf];          // b0042
      if (fabs(dist) > 2.5*pulseRange) err = E_timeout; // 2.5*0.7=1.75
    }
  }
  if (lens == LENS_A) {
    lensA_moving = FALSE; [lensLockA unlock];
  } else {
    lensB_moving = FALSE; [lensLockB unlock];
  }
  [self enableGUI:YES];

  return err;
}

/* ---------------------------------------------------------------- */

- (void)moveLensToTarget
{
  char buf[128];
#if (DEBUG > 0)
  fprintf(stderr,"%s(%.3f,%.3f)\n",PREFUN,lensTargetA,lensTargetB);
#endif

  if (calibrating) return;
  
  sprintf(buf,"%s: %.2f %.2f",PREFUN,lensTargetA,lensTargetB);
  [main_logger append:buf];

  if (fabs(modneg(lensEncoderA-lensTargetA,360)) > pulseRange) {
    dispatch_async(GLOBAL_QUEUE,^(void) {
      int err = [self openAcros];        // b0038
      if (err) {
        [self showError:@"opening Acromag failed" error:err];
      } else {
        err = [self moveLens:LENS_A to:lensTargetA];
        if (err) [self showError:@"moving lens-A failed" error:err]; // now! b0046
        if (!lensB_moving) {             // next correction soon b0037
          @synchronized(self) { next_update = time(NULL)+2; }
          Ysleep(4000);
        }
        [self closeAcros];
      }
    });
  }
  if (fabs(modneg(lensEncoderB-lensTargetB,360)) > pulseRange) {
    dispatch_async(GLOBAL_QUEUE,^(void) {
      int err = [self openAcros];        // b0038
      if (err) {
        [self showError:@"opening Acromag failed" error:err];
      } else {
        err = [self moveLens:LENS_B to:lensTargetB];
        if (err) [self showError:@"moving lens-B failed" error:err]; // now! b0046
        if (!lensA_moving) {             // next correction soon b0037
          @synchronized(self) { next_update = time(NULL)+2; }
          Ysleep(4000);
        }
        [self closeAcros];
      }
    });
  }
}

/* ---------------------------------------------------------------- */
#pragma mark "Run"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (void)run_init
{
  int  err=0;
  const int TCS_PORT=5800;
  BOOL on=NO;
  NSString *host,*str;
#if (DEBUG > 0)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif

  initialized = FALSE;
  
  NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];

  [self enableGUI:NO];
  [self reloadPreferences];

  host = YprefGetString(@DBE_TCS_HOST,@DEF_TCS_HOST,use_shared);
#ifndef SIM_ONLY
  on   = (BOOL)YprefGetInt(DBE_TCS_ONLINE,1,use_shared);
#endif
  tcs = [[MagIO alloc] initWithHost:[host UTF8String] port:TCS_PORT online:on];
  assert(tcs);
  str = [NSString stringWithFormat:@"tcs= %@:%d (%d)",host,TCS_PORT,on];
  [main_logger append:[str UTF8String]];

  if (!tcs.online) {                   //  b0024
    dispatch_async(MAIN_QUEUE,^(void) {
      for (NSTextField *e in @[edit_elevation,edit_rotator]) {
        [e setSelectable:YES]; [e setEditable:YES];
        [e setTarget:self]; [e setAction:@selector(action_telescope:)];
      }
    });
  }
  if (!err) {
    err = [self updateTCS:YES];
    if (err) [app goodbye:@"connecting to TCS failed" error:err];
  }

#ifndef SIM_ONLY
  on = (BOOL)YprefGetInt(DBE_ACRO_A_ONLINE,1,use_shared);
#endif
  host = YprefGetString(@DBE_ACRO_A_HOST,@DEF_ACRO_A_HOST,use_shared);
  acroA = [[AcroIO alloc] initWithHost:[host UTF8String] port:ACRO_PORT online:on];
  assert(acroA);
  str = [NSString stringWithFormat:@"acroA= %@:%d (%d)",host,ACRO_PORT,on];
  [main_logger append:[str UTF8String]];
  
#ifndef SIM_ONLY
  on = (BOOL)YprefGetInt(DBE_ACRO_D_ONLINE,1,use_shared);
#endif
  host = YprefGetString(@DBE_ACRO_D_HOST,@DEF_ACRO_D_HOST,use_shared);
  acroD = [[AcroIO alloc] initWithHost:[host UTF8String] port:ACRO_PORT online:on];
  assert(acroD);
  str = [NSString stringWithFormat:@"acroD= %@:%d (%d)",host,ACRO_PORT,on];
  [main_logger append:[str UTF8String]];

  if (!err) {                          // digital
    if (!acroD.online) Ysleep(500);
    adcState = YprefGetInt(DBE_ADC_STATE,0,use_shared);
    adcLimitOut = (adcState==0) ? 1 : 0;  // Simulator preset
    adcLimitIn  = (adcState==1) ? 1 : 0;
    otherState = YprefGetInt(DBE_OTHER_STATE,0,use_shared);
    otherLimitOut = (otherState==0) ? 1 : 0;
    otherLimitIn  = (otherState==1) ? 1 : 0;
    err = [acroD open];
    if (!err) {
      if (!err) err = [self updateSlides:YES];
      [acroD close];
    }
    if (err) [app goodbye:@"connecting to Acromag/983 failed" error:err];
  }
  
  if (!err) {                          // analog
    if (!acroA.online) Ysleep(500);
    err = [acroA open];
    if (!err) {
      lensEncoderA = YprefGetDouble(DBE_ENCODER_A,0,use_shared);
      if (!err) err = [self updateLens:LENS_A encoder:YES];
      lensEncoderB = YprefGetDouble(DBE_ENCODER_B,0,use_shared);
      if (!err) err = [self updateLens:LENS_B encoder:YES];
      [acroA close];
    }
    if (err) [app goodbye:@"connecting to Acromag/962 failed" error:err];
  }
  
  if (!err) {
    initialized = TRUE;
    [self enableGUI:YES];
    [self updateLensTarget:NO];
    [app updateRunMode]; // b0037
  }

  [pool drain];
}

/* ---------------------------------------------------------------- */

#ifdef SIM_ONLY
- (void)simulator:(int)lens angle:(double)phi sin:(int*)s cos:(int*)c
{
  if (lens == LENS_A) {
    *s = (int)my_round(v0sinA+a0sinA*sin(phi),0);
    *c = (int)my_round(v0cosA+a0cosA*cos(phi),0);
  } else {
    *s = (int)my_round(v0sinB+a0sinB*sin(phi),0);
    *c = (int)my_round(v0cosB+a0cosB*cos(phi),0);
  }
}
#endif

/* --- */

- (void)run_calibration:(NSNumber*)param
{
  int    lens = [param intValue];
  int    s=0,c=0;
#ifdef SIM_ONLY
  double phi=0,dPhi=0.2;
#endif
#if (DEBUG > 0)
  fprintf(stderr,"%s(%d)\n",PREFUN,lens);
#endif
  assert(![NSThread isMainThread]);

  cal_min_sin = cal_min_cos = 0xffff; cal_max_sin = cal_max_cos = 0;
  
  [plotSinCos clearData];
  [plotCircle clearData];
  [plotAngle  clearData];
  dispatch_async(MAIN_QUEUE,^(void) {
    [plotCircle.window orderFront:self];
    [plotSinCos.window orderFront:self];
    [plotAngle.window  orderFront:self];
  });
  calStartTime = walltime();

  int err = [acroA open];
  if (!err) {
    while (calibrating) {
#ifdef SIM_ONLY
      if (!acroA.online) {             // Simulator
        [self simulator:lens angle:phi sin:&s cos:&c];
        phi += dPhi; dPhi = fmax(0.05,0.99*dPhi);
        Ysleep(50);
      }
#endif
      err = [self getEncoder:lens sin:&s cos:&c]; if (err) break;
      cal_min_sin = imin(cal_min_sin,s);
      cal_max_sin = imax(cal_max_sin,s);
      cal_min_cos = imin(cal_min_cos,c);
      cal_max_cos = imax(cal_max_cos,c);
      double v0s = (cal_max_sin+cal_min_sin)/2.0;
      double a0s = (cal_max_sin-cal_min_sin)/2.0;
      double v0c = (cal_max_cos+cal_min_cos)/2.0;
      double a0c = (cal_max_cos-cal_min_cos)/2.0;
      float angle = getAngle(s,v0s,a0s,c,v0c,a0c,cal_phase);
      dispatch_async(MAIN_QUEUE,^(void) { char buf[32];
        [self updatePlot:lens sin:s cos:c angle:angle];
        sprintf(buf,"%d",s); YeditSet(edit_sin,buf);
        sprintf(buf,"%d",cal_min_sin); YeditSet(edit_min_sin,buf);
        sprintf(buf,"%d",cal_max_sin); YeditSet(edit_max_sin,buf);
        sprintf(buf,"%d",c); YeditSet(edit_cos,buf);
        sprintf(buf,"%d",cal_min_cos); YeditSet(edit_min_cos,buf);
        sprintf(buf,"%d",cal_max_cos); YeditSet(edit_max_cos,buf);
        sprintf(buf,"+%.1f",angle); YeditSet(edit_angle,buf);
      });
      Ysleep(200);
    } // endwhile(calibrating)
    [acroA close];
  } // endif(err)
  
  if (err) {
    [self showError:@"calibration failed" error:err];
  } else { __block BOOL saveIt;
    dispatch_sync(MAIN_QUEUE,^(void) { char msg[128],buf[128];
      sprintf(buf,"Save calibration (%s)",LENS_NAME(lens));
      sprintf(msg,"sinMin=%d sinMax=%d\rcosMin=%d cosMax=%d",
              cal_min_sin,cal_max_sin,cal_min_cos,cal_max_cos);
      saveIt = YalertQuestion(buf,msg,0,"Yes","No");
    });
    if (saveIt) { char buf[128];
#ifndef SIM_ONLY // do not overwrite existing calibration in Simulator mode
      YprefPutInt(DBE_MIN_SIN(lens),cal_min_sin,use_shared);
      YprefPutInt(DBE_MAX_SIN(lens),cal_max_sin,use_shared);
      YprefPutInt(DBE_MIN_COS(lens),cal_min_cos,use_shared);
      YprefPutInt(DBE_MAX_COS(lens),cal_max_cos,use_shared);
      YprefPutDouble(DBE_PHASE(lens),cal_phase,use_shared);
#endif
      sprintf(buf,"calibration(%s): sinMin=%d sinMax=%d cosMin=%d cosMax=%d",
        LENS_NAME(lens),cal_min_sin,cal_max_sin,cal_min_cos,cal_max_cos);
      [main_logger append:buf];
      [self reloadCalibration:lens];
    }
    [self updateLens:lens encoder:YES];
  }
  dispatch_sync(MAIN_QUEUE,^(void) {
    [pop_lens setEnabled:YES];
    [but_calibrate setTitle:@"Start Calibration"];
  });

  calibrating = FALSE;
}

/* ---------------------------------------------------------------- */

- (void)run_log:(NSNumber*)param
{
  int    lens = [param intValue];
  int    s=0,c=0,running=0;
#ifdef SIM_ONLY
  double phi=0,dPhi=0.1;
#endif
#if (DEBUG > 0)
  fprintf(stderr,"%s(%d)\n",PREFUN,lens);
#endif
  if (calibrating) return;
  
  @synchronized(self) {
    if (logging) running = 1;
    else         logging = TRUE;
  }
  if (running) return;
  
  [plotSinCos clearData];
  [plotCircle clearData];
  dispatch_async(MAIN_QUEUE,^(void) {
    [plotSinCos.window orderFront:self];
    [plotCircle.window orderFront:self];
    [plotAngle.window  orderOut:self];
  });
  calStartTime = walltime();

  int err = [acroA open];
  if (!err) { float angle; char buf[128];
    while (logging) {
#ifdef SIM_ONLY
      if (!acroA.online) {             // Simulator
        phi += dPhi; dPhi += (drand48()-0.5)/20.0;
        while (fabs(dPhi) > 0.2) dPhi *= 0.99;
        [self simulator:lens angle:phi sin:&s cos:&c];
        Ysleep(50);
      }
#endif
      err = [self getEncoder:lens sin:&s cos:&c]; if (err) break;
      [self updatePlot:lens sin:s cos:c]; 
      if (lens == LENS_A) angle = getAngle(s,v0sinA,a0sinA,c,v0cosA,a0cosA,phaseA);
      else                angle = getAngle(s,v0sinB,a0sinB,c,v0cosB,a0cosB,phaseB);
      sprintf(buf,"+%.1f",angle); YeditSet(edit_angle,buf); // b0036
      if (calibrating) break;
    } // endwhile(logging)
    [acroA close];
  } // endif(err)
  if (err) [self showError:@"logging failed" error:err];
  
  @synchronized(self) { logging = FALSE; }
}

/* ---------------------------------------------------------------- */

@end

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
