/* ---------------------------------------------------------------- */
//
// PlotController.m
//
// Created by Christoph Birk on 2019-02-07
//
// IDEA --> Common
//
/* ---------------------------------------------------------------- */

#define DEBUG           1              // debug level 

/* ---------------------------------------------------------------- */

#import  "PlotController.h"
#import  "Yutils.h"

#if (DEBUG > 0)
#import  "Logger.h"
extern Logger *main_logger;
#endif

#include "utils.h"

/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

@implementation PlotDataPoint

+ (id)pointWithX:(double)x y:(double)y
{
  PlotDataPoint *point = [[PlotDataPoint alloc] initWithX:x y:y];
  return [point autorelease];
}

/* --- */

- (id)initWithX:(double)x y:(double)y
{
  if ((self = [super init]) != nil) {
    _x = x;
    _y = y;
  }
  return self;
}
@end

/* ---------------------------------------------------------------- */
/* PlotController */
/* ---------------------------------------------------------------- */

@implementation PlotController
{
  NSString *key_frame;
  NSArray  *dataArrays;
}

@synthesize view;

/* ---------------------------------------------------------------- */
#pragma mark "Class"
/* ---------------------------------------------------------------- */

#ifndef NDEBUG
+ (void)initialize
{
  if (self != [PlotController class]) return; // prevent sub-classes
#if (DEBUG > 0)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif
}
#endif

/* ---------------------------------------------------------------- */
#pragma mark "Instance"
/* ---------------------------------------------------------------- */

- (id)initWithWindowNib:(NSString*)name context:(int)myid
                                        title:(const char*)tstr
                                        unit:(const char*)ustr
                                        N:(int)nc
                                        colors:(const PlotColor*)cols
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d,%s,%s,%d,%p)\n",PREFUN,myid,tstr,ustr,nc,cols);
#endif

  if ((self = [super initWithWindowNibName:name]) == nil) return nil;

  YwindowTitle([self window],tstr);    // forces lazy window
  key_frame = [[NSString stringWithFormat:@"PlotWindow_%s_%d",tstr,myid] retain];
  [[self window] setFrameUsingName:key_frame];
  [panel_pars setTitle:[NSString stringWithFormat:@"Plot Parameters (%s)",tstr]];
 
  NSMutableArray *array = [NSMutableArray arrayWithCapacity:nc];
  for (int i=0; i<nc; i++) {
    NSMutableArray *data = [[NSMutableArray alloc] initWithCapacity:128];
    [array addObject:data];
    [data release];
  }
  assert([array count] == nc);
  dataArrays = [[NSArray alloc] initWithArray:array];
  assert(view);
  [view setDataArray:dataArrays colors:cols];
  view.unit = @(ustr);

  return self;
}

/* ---------------------------------------------------------------- */

- (void)dealloc
{
  [dataArrays release];
  [key_frame release];
  [super dealloc];
}

/* ---------------------------------------------------------------- */

- (void)windowWillClose:(NSNotification*)notification
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p)\n",PREFUN,notification);
#endif
  YwindowFrameSave(self.window,[key_frame UTF8String]);
}

/* ---------------------------------------------------------------- */

- (void)addData:(int)c x:(double)x y:(double)y redraw:(BOOL)r
{
#if (DEBUG > 2)
  fprintf(stderr,"%s(%d,%f,%.f): %p\n",PREFUN,c,x,y,self);
#endif
  assert(view);
  
  NSMutableArray *array = [dataArrays objectAtIndex:c];
  @synchronized (array) {
    [array addObject:[PlotDataPoint pointWithX:x y:y]];
  }
  if (r) {
    if ([[self window] isVisible]) YviewNeedsDisplay(view);
  }
}


/* ---------------------------------------------------------------- */

- (void)clearData
{
  for (NSMutableArray *array in dataArrays) {
    @synchronized(array) {
      [array removeAllObjects];
    }
  }
#ifndef NDEBUG
  if ([[self window] isVisible]) YviewNeedsDisplay(view);
#endif
}

/* --- */

- (void)removeData:(int)n
{
  for (NSMutableArray *array in dataArrays) {
    @synchronized(array) {
      for (int i=0; i<n; i++) {
        if (array.count < 1) break;
        [array removeObjectAtIndex:0];
      }
    }
  }
#ifndef NDEBUG
  if ([[self window] isVisible]) YviewNeedsDisplay(view);
#endif
}

/* ---------------------------------------------------------------- */

- (void)updateXscale
{
  char   buf[128];
#if (DEBUG > 0)
  fprintf(stderr,"%s: xmode=%d\n",PREFUN,view.scale_xmode);
#endif
  assert([NSThread isMainThread]);

  [popup_xmode selectItemAtIndex:view.scale_xmode];

  switch (view.scale_xmode) {   // 'IB' dependency
  case PLOT_SCALE_MAN:
  case PLOT_SCALE_FULL:
    sprintf(buf,"%.2f",view.scale_xmin); YeditSet(edit_xmin,buf);
    sprintf(buf,"%.2f",view.scale_xmax); YeditSet(edit_xmax,buf);
    [edit_xmin setEnabled:!view.scale_xmode];
    [edit_xmax setEnabled:!view.scale_xmode];
    break;
  }
}

/* ---------------------------------------------------------------- */

- (void)updateYscale
{
  char  buf[128];
#if (DEBUG > 0)
  fprintf(stderr,"%s: xmode=%d\n",PREFUN,view.scale_xmode);
#endif
  assert([NSThread isMainThread]);

  [popup_ymode selectItemAtIndex:view.scale_ymode];

  sprintf(buf,"%.1f",view.scale_ymin); YeditSet(edit_ymin,buf);
  sprintf(buf,"%.1f",view.scale_ymax); YeditSet(edit_ymax,buf);
  [edit_ymin setEnabled:!view.scale_ymode];
  [edit_ymax setEnabled:!view.scale_ymode];
}

/* ---------------------------------------------------------------- */

-(IBAction)action_pars:(id)sender
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif
  assert([NSThread isMainThread]);
  
  if (sender == popup_xmode) {
    view.scale_xmode = (int)[popup_xmode indexOfSelectedItem];
    [self updateXscale];
  } else
  if (sender == popup_ymode) {
    view.scale_ymode = (int)[popup_ymode indexOfSelectedItem];
    [self updateYscale];
  } else
  if (sender == edit_xmin) {
    assert(view.scale_xmode == PLOT_SCALE_MAN);
    view.scale_xmin = [edit_xmin doubleValue];
  } else
  if (sender == edit_xmax) {
    assert(view.scale_xmode == PLOT_SCALE_MAN);
    view.scale_xmax = [edit_xmax doubleValue];
  } else
  if (sender == edit_ymin) {
    assert(view.scale_ymode == PLOT_SCALE_MAN);
    view.scale_ymin = [edit_ymin floatValue];
  } else
  if (sender == edit_ymax) {
    assert(view.scale_ymode == PLOT_SCALE_MAN);
    view.scale_ymax = [edit_ymax floatValue];
  } else
  if (sender == but_apply) {
    [panel_pars orderOut:self];
    if (view.scale_xmode == PLOT_SCALE_MAN) {
      view.scale_xmin = [edit_xmin doubleValue];
      view.scale_xmax = [edit_xmax doubleValue];
    }
    if (view.scale_ymode == PLOT_SCALE_MAN) {
      view.scale_ymin = [edit_ymin doubleValue];
      view.scale_ymax = [edit_ymax doubleValue];
    }
  } else
  if (strstr([[sender title] UTF8String],"Pars")) {
    if (![panel_pars isVisible]) {
      [self updateXscale];
      [self updateYscale];
    }
    [panel_pars makeKeyAndOrderFront:self];
  } else {
    assert(0);
  }
  if (view.scale_xmin > view.scale_xmax) { time_t h;
    h = view.scale_xmin;
    view.scale_xmin = view.scale_xmax;
    view.scale_xmax = h;
  }
  if (view.scale_ymin > view.scale_ymax) { float h;
    h = view.scale_ymin;
    view.scale_ymin = view.scale_ymax;
    view.scale_ymax = h;
  }
  YviewNeedsDisplay(view);
}

/* ---------------------------------------------------------------- */

@end

/* ---------------------------------------------------------------- */
/* GraphView */
/* ---------------------------------------------------------------- */

@implementation PlotView

@synthesize scale_xmode,scale_xmin,scale_xmax;
@synthesize scale_ymode,scale_ymin,scale_ymax;
@synthesize unit;

/* ---------------------------------------------------------------- */

- (id)initWithFrame:(NSRect)frame
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%.0f,%.0f)\n",PREFUN,frame.size.width,frame.size.height);
#endif

  if ((self = [super initWithFrame:frame]) == nil) return nil;

  scale_xmode = PLOT_SCALE_FULL;
  scale_ymode = PLOT_SCALE_FULL;
  
  font_size = 14.0;
  font = CtextGetFont("Helvetica",font_size);

  return self;
}

/* ---------------------------------------------------------------- */

- (void)dealloc 
{
  [dataArrays release];
  if (colors) free((void*)colors);
  [unit release];
  [super dealloc];
}

/* ---------------------------------------------------------------- */

- (void)setDataArray:(NSArray*)array colors:(const PlotColor*)cols
{
  dataArrays = [array retain];
  colors = (PlotColor*)malloc([array count]*sizeof(PlotColor));
  for (int i=0; i<[array count]; i++) { colors[i] = cols[i]; }
}

/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (void)drawRect:(NSRect)rect
{
  int           i,xtics=1,ytics=1,itic,move_flag,nc;
  double        mtic,xtic,ytic,vmin,vmax,vdyn;
  double        t,v,x,y=0,lmargin=0;
  double        xmin,xmax,xdyn;
  char          buf[128];
  CGContextRef  context;
  BOOL          dots_only=TRUE;
  const double  vtic[]={1,2,5};
  const CGFloat rmargin=10,bmargin=18,tmargin=18;
  const CGFloat dashes[]={10,5};
#if (DEBUG > 1)
  fprintf(stderr,"%s([%.0f,%.0f]): %p\n",PREFUN,rect.size.width,rect.size.height,self);
#endif
  assert([NSThread isMainThread]);     // NSView is mainThread only

  if (!dataArrays) return;
  nc = (int)[dataArrays count];
  
  if (scale_xmode) {                   // 'full' scaling
    scale_xmin = 9.9e99; scale_xmax = -9.9e99;
    for (NSArray *array in dataArrays) {
      @synchronized (array) {
        for (PlotDataPoint *p in array) {
          scale_xmin = fmin(scale_xmin,p.x);
          scale_xmax = fmax(scale_xmax,p.x);
        }
      }
    } // endfor(nc)
  }
  if (scale_xmin > scale_xmax) return; // not enough data
#if (DEBUG > 1)
  printf("x_min=%f, x_max=%f (%f)\n",scale_xmin,scale_xmax,scale_xmax-scale_xmin);
#endif
  
  if (scale_ymode) {                   // 'full' scaling
    scale_ymin = 9.9e99; scale_ymax = -9.9e99;
    for (NSArray *array in dataArrays) {
      @synchronized (array) {
        for (PlotDataPoint *p in array) {
          if (p.x < scale_xmin) continue;
          if (p.x > scale_xmax) continue;
          scale_ymin = (float)fmin(scale_ymin,p.y);
          scale_ymax = (float)fmax(scale_ymax,p.y);
        }
      }
    }
    if (scale_ymax-scale_ymin < 0.01) scale_ymax += 0.01;
  }
#if (DEBUG > 1)
  printf("y_min=%.1f, y_max=%.1f\n",scale_ymin,scale_ymax);
#endif

  context = [[NSGraphicsContext currentContext] CGContext];

  float yb = bmargin+self.bounds.origin.y;
  float yt = self.bounds.origin.y+self.bounds.size.height-tmargin;
  float h = yt-yb; if (h < 1) return;

  itic=0; mtic=0.1;                    // design Y-axis
  do {
    ytic = vtic[itic]*mtic;
    vmin = floor(scale_ymin/ytic)*ytic;
    vmax = ceil(scale_ymax/ytic)*ytic;
    if (vmax == vmin) vmax = floor(scale_ymax/ytic+1.00)*ytic;
    if (ytic*ytics >= vmax-vmin) break;
    ytics++;
    if (ytics*30 > h) {
      ytics=1; itic++;                 // (double) bug-fix b0057
      if (itic == (sizeof(vtic)/sizeof(double))) { itic=0; mtic *= 10.0; }
      assert(itic < 3);
    }
  } while (1);
  vdyn = vmax - vmin;
#if (DEBUG > 1)
  printf("vmin=%f, vmax=%f, vdyn=%f\n",vmin,vmax,vdyn);
#endif
  assert(vdyn > 0.0);

  for (i=0; i<=ytics; i++) {           // get required lmargin
    v = vmin + i*ytic;
    if (v < 10.0) sprintf(buf,"%.1f",v);
    else          sprintf(buf,"%.0f",v);
    v = 10 + 0.50*font_size*strlen(buf); if (v > lmargin) lmargin = v;
  }
  float xl = lmargin+self.bounds.origin.x;
  float xr = self.bounds.origin.x+self.bounds.size.width-rmargin;
  float w  = xr-xl; if (w < 1) return;

  itic=0; mtic=0.1;                    // design X-axis
  do {
    xtic = mtic*vtic[itic]; 
    xmin = floor(scale_xmin/xtic)*xtic;
    xmax = ceil(scale_xmax/xtic)*xtic;
    if (xmax == xmin) xmax = floor(scale_xmax/xtic+1.00)*xtic;
    if (xtic*xtics >= xmax-xmin) break;
    xtics++;
    if (xtics*60 > w) {
      xtics=1; itic++;                 // (double) bug-fix b0057
      if (itic == (sizeof(vtic)/sizeof(double))) { itic=0; mtic *= 10.0; }
      assert(itic < 3);
    }
  } while (1);
  xdyn = xmax-xmin;
#if (DEBUG > 1)
  printf("xmin=%f, xmax=%f, xdyn=%f\n",xmin,xmax,xdyn);
#endif
  assert(xdyn > 0);

  CGContextSetRGBStrokeColor(context,0,0,0,1);   // black
  CGContextSetLineWidth(context,2);              // draw frame
  CGContextStrokeRect(context,CGRectMake(xl,yb,w,h));

  CGContextSetTextMatrix(context,CGAffineTransformIdentity);

  CGContextSetLineWidth(context,1);    // draw tics
  CGContextSetLineDash(context,0.0,dashes,sizeof(dashes)/sizeof(CGFloat));
  CGContextBeginPath(context);

  for (i=0; i<=xtics; i++) {           // draw X-legend
    t = i*xtic; x = xl + w*t/xdyn;
    if (xtic < 1.0) sprintf(buf,"%.1f",my_round(xmin+t,1));
    else            sprintf(buf,"%.0f",my_round(xmin+t,0));
    CtextCenterText(buf,font,context,x-((i==xtics) ? 0.6*font_size : 0),yb-0.6*font_size);
    if ((i==0) || (i == xtics)) continue;
    CGContextMoveToPoint(context,x,yb);
    CGContextAddLineToPoint(context,x,yt);
  }

  for (i=0; i<=ytics; i++) {           // draw Y-legend
    v = i*ytic; y = yb + h*v/vdyn;
    v += vmin;
    if (fabs(v) < 10.0) sprintf(buf,"%.1f",v);
    else                sprintf(buf,"%.0f",v);
    CtextRightText(buf,font,context,xl-0.4*font_size,y+0.25*font_size);
    if ((i == 0) || (i == ytics)) continue;
    CGContextMoveToPoint(context,xl,y);
    CGContextAddLineToPoint(context,xr,y);
  }
  if ([unit compare:@""] != NSOrderedSame) {
    sprintf(buf,"[%s]",[unit UTF8String]);
    CtextRightText(buf,font,context,xl-0.6*font_size,y-1.0*font_size);
  }
  CGContextDrawPath(context,kCGPathStroke);  // draw dashed lines

  /* draw sensor data --------------------------------------------- */

  CGContextSetLineDash(context,0.0,NULL,0);  // solid line
  CGContextSetLineWidth(context,(self.bounds.size.height > 260) ? 2 : 1);
  CGContextClipToRect(context,CGRectMake(xl+1,yb+1,w-2,h-2));
  for (i=0; i<nc; i++) {               // loop over channels
    CGContextSetRGBStrokeColor(context,colors[i].r,colors[i].g,colors[i].b,1);
    CGContextSetRGBFillColor(context,colors[i].r,colors[i].g,colors[i].b,1);
    if (!dots_only) CGContextBeginPath(context);
    move_flag = 1;  
    NSArray *array = [dataArrays objectAtIndex:i];
    @synchronized (array) {
      for (PlotDataPoint *point in array) {
        t = point.x;
        if ((t < xmin) || (t > xmax)) { move_flag=1; continue; }
        v = point.y;
        x = xl + w*(t-xmin)/xdyn;
        y = yb + h*(v-vmin)/vdyn;
        if (dots_only) {
          CGContextBeginPath(context);
          CGContextAddArc(context,x,y,3,0,2*3.14159,0);
          CGContextFillPath(context);
        } else {
          if (move_flag) CGContextMoveToPoint(context,x,y);
          else           CGContextAddLineToPoint(context,x,y);
          move_flag = 0;
        }
      }
    }
    if (!dots_only) CGContextDrawPath(context,kCGPathStroke);  // draw line
  }
}

/* ---------------------------------------------------------------- */

@end

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
